
   ' =====================================================================
   ' Projekt   Angry Bird Wing Control
   ' Autor     Robert Link 2021
   ' Bascom    2.0.8.4
   ' Version   3.0.0
   '
   ' Beschreibung:
   ' Ein Servo bewegt sich hin und her und ermglicht so ein "Flattern" der Flgel. In Abhngigkeit des eingelesenen RC Kanals
   ' kann die Flattergeschwindigkeit reguliert werden, das Servo angehalten oder in Mittelstellung gebracht werden.
   ' Funktion:
   ' RC Kanal befindet sich im unteren Viertel (0-25%): Servo steht in Mittelstellung (fr Start und Landung)
   ' RC Kanal befindet sich im zweiten Viertel (25-50%): Servo bleibt in aktueller Stellung stehen
   ' RC Kanal befindet sich im obererer Hlfte (50-100%): Servo bewegt sich hin und her. Je weiter der RC Kanal Richtung 100% gebracht wird, umso schneller werden die Bewegungen
   ' Servo-Neutralstellung kann mittels Poti eingestellt werden
   ' Bei fehlendem RC-Signal wird nach einer Sekunde mittels Watchdog ein Reset durchgefhrt (Servo geht in Mittelstellung) ACHTUNG: FUNKTIONIERT NUR MIT EINGEBAUTEM R1 !

   ' RC Signalerfassung und Servoausgabe erfolgt jeweils mit einem eigenen Timer
   '
   ' RC Eingang 1 an Pin 7 (PB2, INT0)
   ' Servo an Pin 6 (PB1)
   ' Poti (10k) an Pin 2 (ADC3)
   ' Kontroll-LED an Pin 3 (PB4), optional
   '
   ' Die Nutzung des Codes (auch auszugsweise) ist ausschlielich fr nicht
   ' kommerziellen Nutzung freigegeben (Namensnennung des Autors)!
   ' Der Code ist experimentell und die Benutzung erfolgt auf eigene Gefahr!
   ' Der Sourcecode darf fr den Eigenbedarf beliebig gendert werden.
   ' Kontakt: www.bascomforum.de "R2D2 Bastler"
   '
   ' =====================================================================


   $Regfile = "attiny45.dat"
   $Crystal = 8000000                                       ' FuseBit CKDIV8 deaktivieren
   $HWstack = 40
   $SWstack = 40
   $Framesize = 60

'  --------------------------------------------------------
'  Globale Definitionen
'  --------------------------------------------------------
   Const TRUE = 1
   Const FALSE = 0
   Config SubMode = New

'  --------------------------------------------------------
'  Timer 0 konfigurieren (CTC-Mode) fr Servopulse-Ausgabe
'  --------------------------------------------------------
   Config Timer0 = Timer , Prescale = 64 , Clear_Timer = 1  ' Timer Tick = 8us, 125 Ticks entsprechen 1ms, 250 Ticks entsprechen 2ms
   OCR0A = 187                                              ' Zu Beginn Compare Match Wert auf 187 (entspricht 1,5ms) stellen
   Stop Timer0                                              ' Timer0 erst unmittelbar vor Programmbeginn starten lassen, da es sonst beim WatchDog-Neustart zu Servozucken kommt
   On OC0A ServoPulseOutput Nosave                          ' Register werden manuel in der ISR gesichert
   Enable OC0A                                              ' OC0A-Interrupt zulassen

'  --------------------------------------------------------
'  Timer 1 konfigurieren fr Erfassung des RC Signals, Prescale 128
'  --------------------------------------------------------
   Config Timer1 = Timer , Prescale = 128                   ' Timer Tick = 16us
   ' Info:
   ' "Config Timer1 = Timer , Prescale = 128" funktionierte bei frheren
   ' Bascom-Versionen beim Attiny 25/45/85 leider nicht
   ' "TCCR1 = &B00001000" ersetzte damals "Config Timer1 = Timer , Prescale = 128"
   ' 62,5 Ticks entsprechen 1ms
   ' 94 Ticks entsprechen 1,5ms
   ' 125 Ticks entsprechen 2ms
   ' Somit keine Timerberlaufprobleme mehr bei langen RC Eingangssignalen ber 2ms

'  --------------------------------------------------------
'  ISR fr Erfassung des RC Signals konfigurieren
'  --------------------------------------------------------
   Config Int0 = Change                                     ' Beim Flankenwechsel an PB2/INT0 (RC Eingang) Int0 auslsen und in die Subroutine springen
   On Int0 RC_ISR Nosave                                    ' Register werden manuel in der ISR gesichert
   Enable Int0

'  --------------------------------------------------------
'  ADC konfigurieren (Einlesen des Potis)
'  --------------------------------------------------------
   Config Adc = Single , Prescaler = Auto
   Start Adc

'  --------------------------------------------------------
'  Konstanten definieren
'  --------------------------------------------------------
   'RC Input Grenzwerte (Timer1 Prescale 128)
   Const RC_in_min = 62                                     ' Minimaler RC-Wert fr die Berechnungen (min Wert 62 fr 1ms)
   Const RC_in_max = 125                                    ' Maximaler RC-Wert fr die Berechnungen (max Wert 125 fr 2ms)
   Const Threshold_Servo_Stop = 78                          ' Oberhalb dieses RC-Wertes (1,25ms = 78) bleibt das Servo stehen, wo es sich gerade befindet, unterhalb dieses RC-Wertes geht Servo in Mittelstellung
   Const Threshold_Servo_Flutter = 94                       ' Oberhalb dieses RC-Wertes (1,5ms = 94) bewegt sich das Servo hin und her

   'Servo Output Werte (Timer0 Prescale 64)
   Const Servo_out_min = 125                                ' Minimaler Servowert (125 fr 1ms)
   Const Servo_out_center_default = 187                     ' Servowert fr Mittelstellung, wird noch durch Potistellung feinjustiert (187 fr 1,5ms)
   Const Servo_out_max = 250                                ' Maximaler Servowert (250 fr 2ms)

'  --------------------------------------------------------
'  Variablen definieren
'  --------------------------------------------------------
   'Variablen zur Erfassung des RC-Signals
   Dim RC_Signal_Start As Byte                              ' Timer1 Wert zum Zeitpunkt der steigenden RC-Signal-Flanke
   Dim RC_Signal_End As Byte                                ' Timer1 Wert zum Zeitpunkt der fallenden RC-Signal-Flanke
   Dim Raw_Pulse_Length as Byte                             ' Berechnete RC-Pulslnge (ohne Limit)
   Dim RC_Pulse_Length As Byte                              ' Berechnete RC-Pulslnge (final, mit Limit)

   'Variablen fr Berechnungen des Flgelflatterns
   Dim Up_Down_Flag As Byte                                 ' Flag fr die Servo-Bewegungsrichtung
   Dim FlutterspeedWord as Word                             ' Flattergeschwindigkeit
   Dim Temp_Byte As Byte                                    ' Hilfvariable

   'Variable fr Servoausgabe
   Dim Servo_Value As Byte                                  ' Compare-Wert (Timer0) fr die Servoausgabe

'  --------------------------------------------------------
'  Ein- und Ausgang festlegen
'  --------------------------------------------------------
   LED Alias PortB.4
   Ddrb = &B00010010                                        'PB1 und PB4 wird Ausgang, restlicher PortB bleibt Eingang

'  -----------------------------------------
'  Include Dateien
'  -----------------------------------------
   $Include "Include\Statemachine.inc"                      ' Statemaschine
   $Include "Include\Utilities.inc"                         ' Sonstige Routinen

'  --------------------------------------------------------
'  Watchdog konfigurieren (1sec) und starten
'  --------------------------------------------------------
   Config Watchdog = 1024
   Start Watchdog                                           ' Watchdog starten, Reset erfogt bei anliegendem RC Signal in der ISR "RC_ISR"

'  --------------------------------------------------------
'  Startwerte setzen
'  --------------------------------------------------------
   Call setState(ST_WING_CENTERED)                          ' Flgel in Neutralstellung
   RC_Pulse_Length = RC_in_min                              ' RC Wert vorsichtshalber zu Beginn auf minimum stellen
   Servo_Value = Calc_Servo_Center()                        ' Poti abfragen und berechnete Mittelstellung fr Servo setzen
   Waitms 200
   Start Timer0                                             ' Timer0 (Servosignalausgabe) erst unmittelbar vor Programmbeginn starten lassen, da es sonst beim WatchDog-Neustart zu Servozucken kommt
   Enable Interrupts                                        ' Interrupts zuslassen


'  -----------------------------------------
'  Hauptprogramm
'  -----------------------------------------
   Do
      Select Case getState()

      '  ----------------------------------------
         Case ST_WING_CENTERED                              ' Flgel in Neutralstellung
      '  ----------------------------------------
            If IsStateChanged() = TRUE then                 ' Einmalige Aktion bei Statewechsel
               Reset LED                                    ' optionale LED
            End If

            Servo_Value = Calc_Servo_Center()               ' Poti abfragen und berechnete Mittelstellung fr Servo setzen
            Call Check_RC()                                 ' Prfen, ob neuer RC-Puls gemessen wurde
            If RC_Pulse_Length > Threshold_Servo_Flutter then Call setState(ST_WING_FLUTTER)       ' Wenn die Schaltschwelle berschritten wurde -> Flgelflattern beginnen


      '  ----------------------------------------
         Case ST_WING_FLUTTER                               ' Flgel flattern
      '  ----------------------------------------
            If IsStateChanged() = TRUE then                 ' Einmalige Aktion bei Statewechsel
               Set LED                                      ' optionale LED
            End If

            ' Hier wird geflatter
            Select Case Up_Down_Flag                        ' Hoch- oder Runterflattern?
               Case 0                                       ' Hochzhlen bis Obergrenze
                  Incr Servo_Value                          ' Servo Wert erhhen
                  If Servo_Value > Servo_out_max Then       ' Wenn Obergrenze erreicht
                     Servo_Value = Servo_out_max            ' Wert begrenzen
                     Up_Down_Flag = 1                       ' Marker fr Runterzhlen setzen
                  End If
               Case else                                    ' Runterzhlen bis Untergrenze
                  Decr Servo_Value                          ' Servo Wert verringern
                  If Servo_Value < Servo_out_min Then       ' Wenn Untergrenze erreicht
                     Servo_Value = Servo_out_min            ' Wert begrenzen
                     Up_Down_Flag = 0                       ' Marker fr Raufzhlen setzen
                  End If
            End Select

            ' Flattergeschwindigkeit berechnen
            ' Info: RC_Pulse_Length kann hier nur zwischen Threshold_Servo_Flutter (94) und RC_in_max liegen (125)
            Temp_Byte = RC_Pulse_Length - Threshold_Servo_Flutter       ' Werte 0 bis 31
            Temp_Byte = 37 - Temp_Byte                      ' Werte anheben und Reihenfolge umkehren, Werte 37 bis 7
            FlutterspeedWord = Temp_Byte * 333              ' Werte 12321 bis 2331
            Waitus FlutterspeedWord                         ' Wartezeit bis zur nchsten Servobewegung

            Call Check_RC()                                 ' Prfen, ob neuer RC-Puls gemessen wurde
            If RC_Pulse_Length < Threshold_Servo_Flutter then Call setState(ST_WING_STOP)       ' Wenn die Schaltschwelle unterschritten wurde -> Flgelflattern beenden


      '  ----------------------------------------
         Case ST_WING_STOP                                  ' Flgel bleiben dort stehen, wo sie gerade sind
      '  ----------------------------------------
            If IsStateChanged() = TRUE then                 ' Einmalige Aktion bei Statewechsel
               Reset LED                                    ' optionale LED
            End If

            Call Check_RC()                                 ' Prfen, ob neuer RC-Puls gemessen wurde
            If RC_Pulse_Length < Threshold_Servo_Stop then Call setState(ST_WING_CENTERED)       ' Wenn die Schaltschwelle unterschritten wurde -> Flgel in Neutralstellung
            If RC_Pulse_Length > Threshold_Servo_Flutter then Call setState(ST_WING_FLUTTER)       ' Wenn die Schaltschwelle berschritten wurde - > Flgelflattern beginnen

      End Select

   Loop
   End




'  --------------------------------------------------------
'  RC ISR
'  Wird bei jedem Flankenwechsel am INT0 Pin angesprungen
'  und erfasst den aktuellen Wert von Timer1
'  --------------------------------------------------------
   RC_ISR:
         !push r16                                          ' Register auf Stack sichern
         !in r16,sreg                                       ' Statusregister holen und halten
         !push r17                                          ' Register auf Stack sichern
         !push r18                                          ' Register auf Stack sichern
         !in r17 , tcnt1                                    ' Timer1 Wert holen
         !sbis pinb , 2                                     ' Skip next Instr if PINBx = 1 (Pin PB2 ist beim Attiny85 der INT0 Pin)
         !rjmp Pulse_End1                                   ' Springe zu Pulse_End
         !sts {RC_Signal_Start} , r17                       ' Speichere Timer1 nach RC_Signal_Start
         !rjmp ISR_End1                                     ' Springe zum Ende
      Pulse_End1:
         !lds r18 , {RC_Signal_Start}                       ' Hole RC_Signal_Start Wert
         !SUB r17, r18                                      ' Subtrahiere RC_Signal_Start von RC_Signal_End
         !sts {Raw_Pulse_Length} , r17                      ' Speichere berechnete Pulslaenge in Raw_Pulse_Length
      ISR_End1:
         !pop r18                                           ' Register vom Stack zurckholen
         !pop r17                                           ' Register vom Stack zurckholen
         !Out Sreg , R16                                    ' Statusregister zurckspeichern
         !pop r16                                           ' Register vom Stack zurckholen
         !Wdr                                               ' Watchdog resetten
   Return
'(
   ' Bascom Code (ohne "NoSave", ohne ASM)
   RC_ISR:
      If PINB.2 = 1 then                                    ' Hat steigende RC-Signal-Flanke zum Interrupt gefhrt?                                       '
         RC_Signal_Start = Timer1                           ' Zhlerstand merken
         Else                                               ' oder hat fallende RC-Signal-Flanke zum Interrupt gefhrt?
         RC_Signal_End = Timer1                             ' Zhlerstand merken
         Raw_Pulse_Length = RC_Signal_End - RC_Signal_Start ' Pulslnge berechnen
      End If
      Reset Watchdog                                        ' Watchdog resetten
   Return
')

'  --------------------------------------------------------
'  ISR fr die Ausgabe des Servosignals
'  Wird bei jedem Timer0 berlauf angesprungen
'  Prescale und Vorladewert fr Timer0 werden dabei jedes
'  mal neu gesetzt
'  --------------------------------------------------------

   ServoPulseOutput:
         !push r16                                          ' Register auf Stack sichern
         !in r16,sreg                                       ' Statusregister holen und halten
         !push r17                                          ' Register auf Stack sichern
         !sbic portB , 1                                    ' Skip next instr. wenn PORTB.1 = 0 ist
         !rjmp Start_Low_Phase                              ' Springe zum LOW-Setzen des Servosignals
         !sbi portB , 1                                     ' Setze Servosignal Pin auf HIGH
         !ldi r17 , &B00000011                              ' CS00 und CS01 fr Prescaler 64 in Timer0 vorbereiten
         !Out Tccr0b , R17                                  ' Timer0 luft mit 8s Auflsung
         !lds r17 , {Servo_Value}                           ' Hole aufbereiteten Servowert fr Timer0
         !Out OCR0A , R17                                   ' Neuen Wert ins Compareregister von Timer0 schreiben
         !rjmp End_ISR                                      ' Springe zum Ende der ISR
      Start_Low_Phase:                                      ' Beginn der Low-Phase (ca 19ms)
         !cbi portB , 1                                     ' Setze Servosignal Pin auf LOW
         !ldi r17 , &B00000101                              ' CS00 und CS02 fr Prescaler 1024 in Timer0 vorbereiten
         !Out Tccr0b , R17                                  ' Timer0 luft mit 128s Auflsung
         !ldi r17 , 149-1                                   ' Wert fr ca 19ms (128s*149=19,072ms) bis zum nchsten Comparematch bei Prescaler 1024
         !Out OCR0A , R17                                   ' Neuen Wert ins Compareregister von Timer0 schreiben
      End_ISR:
         !pop r17                                           ' Register vom Stack zurckholen
         !Out Sreg , R16                                    ' Statusregister zurckspeichern
         !pop r16                                           ' Register vom Stack zurckholen
   Return
'(
   ' Bascom Code (ohne "NoSave", ohne ASM)
   ServoPulseOutput:
      If PORTB.1 = 1 then                                   ' Wurde gerade die High-Phase des Servosignals ausgegeben?
         PORTB.1 = 0                                        ' Dann die Low-Phase starten
         Tccr0b = &B00000101                                ' Prescaler von Timer0 auf 1024 setzen (128s Auflsung)
         OCR0A = 149 - 1                                    ' Neuen Wert ins Compareregister von Timer0 schreiben (128s*149=19,072ms)
         Else                                               ' Wurde gerade die Low-Phase des Servosignals ausgegeben?
         PORTB.1 = 1                                        ' Dann die High-Phase starten
         Tccr0b = &B00000011                                ' Prescaler von Timer0 auf 64 setzen (8s Auflsung)
         OCR0A = Servo_Value                                ' Servo-Wert ins Compareregister von Timer0 schreiben
      End If
   Return
')